searchData={"items":[{"type":"module","title":"Charms","doc":"Documentation for `Charms`.\n\n## `defm` and intrinsic\nThere are two ways to define a function with `defm/2` or implement callbacks of `Charms.Intrinsic` behavior. The `defm/2` is a macro that generates a function definition in Charm. The intrinsic is a behavior that generates a function definition in MLIR.\n\nThe intrinsic is more flexible than `defm` because:\n- Intrinsic can be variadic and its argument can be anything\n- Intrinsic is suitable for the cases where directly writing or generating MLIR is more ideal\n- An intrinsic should be responsible for its type check while the Charm’s type system is responsible for function call’s type check\n\nThe `defm` is more suitable for simple functions because it is designed to be as close to vanilla Elixir as possible. As a rule of thumb, use `defm` for simple functions and intrinsic for complex functions or higher-order(generic) function with type as argument.\n\n## `defm`'s differences from `Beaver.>>>/2` op expressions\n- In `Beaver.>>>/2`, MLIR code are expected to mixed with regular Elixir code. While in `defm/2`, there is only Elixir code (a subset of Elixir, to be more precise).\n- In `defm/2`, the extension of the compiler happens at the function level (define your intrinsics or `defm/2`s), while in `Beaver.>>>/2`, the extension happens at the op level (define your op expression).\n- In `Beaver.>>>/2` the management of MLIR context and other resources are done by the user, while in `defm/2`, the management of resources are done by the `Charms` compiler.\n- In `defm/2`, there is expected to be extra verifications built-in to the `Charms` compiler (both syntax and types), while in `Beaver.>>>/2`, there is none.","ref":"Charms.html"},{"type":"module","title":"Caveats and limitations - Charms","doc":"- We need a explicit `call` in function call because the `::` special form has a parser priority  that is too low so a `call` macro is introduced to ensure proper scope.\n- Being variadic, intrinsic must be called with the module name. `import` doesn't work with intrinsic functions while `alias` is supported.","ref":"Charms.html#module-caveats-and-limitations"},{"type":"module","title":"Glossary of modules - Charms","doc":"- `Charms`: the top level macros `defm` and `use Charms`\n- `Charms.Defm`: the `defm` DSL syntax and special forms\n- `Charms.Defm.Definition`: functions to define and compile `defm` functions to MLIR\n- `Charms.Intrinsic`: the behavior used to define and compile intrinsic functions","ref":"Charms.html#module-glossary-of-modules"},{"type":"macro","title":"Charms.defm/2","doc":"define a function that can be JIT compiled","ref":"Charms.html#defm/2"},{"type":"module","title":"Charms.Defm","doc":"Charms.Defm is a DSL for defining functions that can be JIT-compiled.\n\nThis module defines the `defm` DSL syntax as macros and special forms.","ref":"Charms.Defm.html"},{"type":"module","title":"Extending the `defm` - Charms.Defm","doc":"- use `beaver`'s DSL to define intrinsics which can be called in the function body of a `defm`\n- use `defm` to define functions that can be JIT-compiled","ref":"Charms.Defm.html#module-extending-the-defm"},{"type":"macro","title":"Charms.Defm.call/1","doc":"call a local function with return","ref":"Charms.Defm.html#call/1"},{"type":"macro","title":"Charms.Defm.call/2","doc":"call a function defined in another `Charms` module with return","ref":"Charms.Defm.html#call/2"},{"type":"macro","title":"Charms.Defm.cond_br/2","doc":"`cond` expression requires identical types for both branches","ref":"Charms.Defm.html#cond_br/2"},{"type":"macro","title":"Charms.Defm.const/1","doc":"syntax sugar to create an MLIR value from an Elixir value","ref":"Charms.Defm.html#const/1"},{"type":"macro","title":"Charms.Defm.for_loop/2","doc":"for loop","ref":"Charms.Defm.html#for_loop/2"},{"type":"macro","title":"Charms.Defm.op/1","doc":"create an MLIR operation","ref":"Charms.Defm.html#op/1"},{"type":"macro","title":"Charms.Defm.value/1","doc":"create an MLIR operation and return the result value(s)","ref":"Charms.Defm.html#value/1"},{"type":"macro","title":"Charms.Defm.while/2","doc":"while loop","ref":"Charms.Defm.html#while/2"},{"type":"module","title":"Charms.Defm.Definition","doc":"Charms.Defm.Definition provides functions to define and compile `defm` functions that can be JIT/AOT compiled to native targets.","ref":"Charms.Defm.Definition.html"},{"type":"module","title":"The compilation process - Charms.Defm.Definition","doc":"The principle of the compilation process:\n- convert the Elixir code into MLIR IR while preserving the Elixir semantics.\n- work with Elixir's code analysis infrastructure and LSP-like tools.\n\nWith this in mind, the compilation process is divided into these steps:\n1. Convert the Elixir code into MLIR IR, then optimize the IR.\n2. Analyze the AST and IR and generate proper Elixir side effects.\n3. Finally convert the IR into a bytecode representation and persist it as an Elixir string.\n\n```mermaid\ngraph TD\n  A[Elixir Code] -->|Step 1: Convert to MLIR IR| B[MLIR IR]\n  B -->|Optimize IR| C[Optimized MLIR IR]\n  C -->|Step 2: Analyze AST and IR| D[Generate Elixir Side Effects]\n  D --> E[Elixir Function Definitions]\n  D --> F[Remote Function Calls]\n  C -->|Step 3: Convert IR to Bytecode| G[Bytecode Representation]\n  G -->|Persist as Elixir String| H[Persisted Bytecode]\n  H -->|Consumed by JIT/AOT Engine| I[Further Optimization]\n  I --> J[Native Target Code]\n```\n\nThe optimization before JIT/AOT should focus on:\n  - Complete the semantics which usually absent in the Elixir code, such as the return type of a function.\n  - Fill the IR with Elixir's debug information, including code location and function arity.\nElixir side effects include:\n  - Elixir function definitions as placeholder\n  - Remote function calls to build the module dependencies\nThe IR will be later consumed and further optimized by a JIT/AOT engine. It should be as target independent as possible.","ref":"Charms.Defm.Definition.html#module-the-compilation-process"},{"type":"function","title":"Charms.Defm.Definition.compile/1","doc":"Compile definitions into MLIR module.\nCompile definitions into an MLIR module.","ref":"Charms.Defm.Definition.html#compile/1"},{"type":"function","title":"Partial and lazy compilation - Charms.Defm.Definition.compile/1","doc":"When compiling `defm`, Charms performs preprocessing before the definition is fully translated to the target language or IR. Notable preprocessing steps include:\n- Extracting the return type of the `defm` definition, and wrapping it as an anonymous function to be called on-demand at the invocation site.\n- Determine the MLIR ops available in the definition.","ref":"Charms.Defm.Definition.html#compile/1-partial-and-lazy-compilation"},{"type":"function","title":"Charms.Defm.Definition.declare/3","doc":"Declare a function that can be JIT compiled and generate Elixir invoker function.\nDeclare a function that can be JIT compiled and generate Elixir invoker function.\n\nThe call signature will be decomposed and transformed into a normalized form.","ref":"Charms.Defm.Definition.html#declare/3"},{"type":"module","title":"Charms.Defm.Expander","doc":"Expander is a module of functions to compile Elixir AST to MLIR.\n\nIn MLIR, the process of creating IR from source or AST is usually called \"import\". While in our case, we are following Elixir's convention to call it \"expansion\". While we are not just expanding Elixir's AST tree, in the end one Elixir function definition `defm` will be compiled to one MLIR `func.func`.","ref":"Charms.Defm.Expander.html"},{"type":"module","title":"Working with Elixir's parallel compiler - Charms.Defm.Expander","doc":"Charms will try to work with Elixir's parallel compiler in the most well-integrated way. This allows us to compile multiple modules at the same time, which can speed up the compilation process. Another benefit is that all `defm` functions can also be executed at compile time, just like vanilla Elixir functions defined by `def`. Behind the scenes, `func.func` generated from `defm` will also be used as reference to infer and check types of a remote function calls when compiling another module.\nTo take deep dive, Elixir has [an official blog post about parallel compiler](https://elixir-lang.org/blog/2012/04/24/a-peek-inside-elixir-s-parallel-compiler/).","ref":"Charms.Defm.Expander.html#module-working-with-elixir-s-parallel-compiler"},{"type":"module","title":"Cycle reference - Charms.Defm.Expander","doc":"Charms will disallow cycle reference. If module A calls module B, then module B cannot call module A. This is to prevent infinite recursion and stream line compile times features. Although this is technically possible as long as function calls and function definitions are following same signature (like the separated complication of C/C++ files and linkage), we still make this trade-off for simplicity.","ref":"Charms.Defm.Expander.html#module-cycle-reference"},{"type":"module","title":"Compile an AST without enough type information - Charms.Defm.Expander","doc":"Being an dynamic language, it is possible for Elixir to have an AST that is valid in Elixir but not in MLIR. For example, Elixir allows to define a serial of nested function calls without any return type, like `a(b(c()))`. In Elixir everything is expression so a function call is assumed to always return a value while in MLIR it is possible to have a function call that does not return anything (equivalent to a void function in C). In this case, Charms will generate a `ub.poison` operation with result type `none`. If the result value of created `ub.poison` op will never be used, nothing will happen. If used, it will raise an error in later verification or passes. This is meant to allow Elixir code to work with the AST with interest only on Elixir semantic keep going without interruption as much as possible, and limit the error information to the type level, instead of leaking it to the syntax level.","ref":"Charms.Defm.Expander.html#module-compile-an-ast-without-enough-type-information"},{"type":"module","title":"Return type convention - Charms.Defm.Expander","doc":"MLIR allows multiple values as a function return, while Elixir only allows one. To keep the convention, internally Charms will always use a list to store function's return type. If the function has only one return, it will be a list with one element. If the function has no return, it will be an empty list. This should streamline the transformation and pattern matching on the function signature.","ref":"Charms.Defm.Expander.html#module-return-type-convention"},{"type":"function","title":"Charms.Defm.Expander.decompose_call_signature/1","doc":"Decomposes a call into the call part and return type.","ref":"Charms.Defm.Expander.html#decompose_call_signature/1"},{"type":"function","title":"Charms.Defm.Expander.expand/2","doc":"","ref":"Charms.Defm.Expander.html#expand/2"},{"type":"function","title":"Charms.Defm.Expander.expand_to_mlir/3","doc":"Expand an AST into MLIR.","ref":"Charms.Defm.Expander.html#expand_to_mlir/3"},{"type":"module","title":"Charms.Env","doc":"Intrinsic module for BEAM environment's type.","ref":"Charms.Env.html"},{"type":"behaviour","title":"Charms.Intrinsic","doc":"Behaviour to define intrinsic functions.","ref":"Charms.Intrinsic.html"},{"type":"macro","title":"Charms.Intrinsic.defintrinsic/1","doc":"","ref":"Charms.Intrinsic.html#defintrinsic/1"},{"type":"callback","title":"Charms.Intrinsic.handle_intrinsic/3","doc":"Callback to implement an intrinsic.\n\nHaving different return types, there are two kinds of intrinsic functions:\n- Regular: returns a MLIR value or operation.\n- Higher-order: returns a function that returns a MLIR value or operation.","ref":"Charms.Intrinsic.html#c:handle_intrinsic/3"},{"type":"callback","title":"More on higher-order intrinsic - Charms.Intrinsic.handle_intrinsic/3","doc":"Higher-order intrinsic function can be variadic, which means it a list will be passed as arguments.","ref":"Charms.Intrinsic.html#c:handle_intrinsic/3-more-on-higher-order-intrinsic"},{"type":"type","title":"Charms.Intrinsic.intrinsic_return/0","doc":"","ref":"Charms.Intrinsic.html#t:intrinsic_return/0"},{"type":"type","title":"Charms.Intrinsic.ir_return/0","doc":"","ref":"Charms.Intrinsic.html#t:ir_return/0"},{"type":"type","title":"Charms.Intrinsic.opt/0","doc":"","ref":"Charms.Intrinsic.html#t:opt/0"},{"type":"type","title":"Charms.Intrinsic.opts/0","doc":"","ref":"Charms.Intrinsic.html#t:opts/0"},{"type":"module","title":"Charms.JIT","doc":"","ref":"Charms.JIT.html"},{"type":"function","title":"Charms.JIT.destroy/1","doc":"","ref":"Charms.JIT.html#destroy/1"},{"type":"function","title":"Charms.JIT.engine/1","doc":"Returns the JIT engine for the given module.","ref":"Charms.JIT.html#engine/1"},{"type":"function","title":"Charms.JIT.init/2","doc":"","ref":"Charms.JIT.html#init/2"},{"type":"function","title":"Charms.JIT.invoke/2","doc":"","ref":"Charms.JIT.html#invoke/2"},{"type":"module","title":"Charms.Pointer","doc":"Intrinsic module to work with pointers.","ref":"Charms.Pointer.html"},{"type":"module","title":"Charms.Prelude","doc":"Intrinsic module to define essential functions provided by Charms.","ref":"Charms.Prelude.html"},{"type":"function","title":"Charms.Prelude.handle_intrinsic/3","doc":"","ref":"Charms.Prelude.html#handle_intrinsic/3"},{"type":"module","title":"Charms.SIMD","doc":"Intrinsic module for SIMD types.","ref":"Charms.SIMD.html"},{"type":"module","title":"Charms.Term","doc":"Intrinsic module for SIMD type.","ref":"Charms.Term.html"},{"type":"module","title":"SortUtil","doc":"","ref":"SortUtil.html"},{"type":"function","title":"SortUtil.copy_terms/2","doc":"","ref":"SortUtil.html#copy_terms/2"},{"type":"extras","title":"Tutorial: Programming with Charms","doc":"# Tutorial: Programming with Charms\n\n```elixir\nMix.install([\n  {:charms, \"~> 0.1.1\"}\n])\n```","ref":"programming-with-charms.html"},{"type":"extras","title":"Introduction - Tutorial: Programming with Charms","doc":"Charms is an Elixir compiler that allows you to compile a subset of Elixir to native targets. It provides a Domain-Specific Language (DSL) for defining functions that can be JIT-compiled and executed alongside the BEAM virtual machine. In this tutorial, we will explore how to use Charms to write native code and run it.","ref":"programming-with-charms.html#introduction"},{"type":"extras","title":"Setting Up Your Module - Tutorial: Programming with Charms","doc":"To start using Charms, define a module and include the Charms DSL:\n\n```elixir\ndefmodule Charming do\n  use Charms\nend\n```\n\nThis sets up your module to leverage Charms and imports commonly used modules.","ref":"programming-with-charms.html#setting-up-your-module"},{"type":"extras","title":"Defining Functions with `defm` - Tutorial: Programming with Charms","doc":"Instead of using the standard Elixir `def`, utilize `defm` to define functions that can be JIT-compiled:\n\n```elixir\ndefmodule HaveFun do\n  use Charms\n  alias Charms.Term\n\n  defm my_fun(env, arg1 :: Term.t(), arg2 :: Term.t()) :: Term.t() do\n    func.return(arg2)\n  end\nend\n\nHaveFun.my_fun(1, :hello)\n```\n\nNote the type annotations here. The first argument, `env`, represents the BEAM environment type (`env :: Charms.Env.t()`), and the types `Term.t()` specify Erlang terms in C.","ref":"programming-with-charms.html#defining-functions-with-defm"},{"type":"extras","title":"Calling Other Functions - Tutorial: Programming with Charms","doc":"To call other functions defined with `defm`, use the `call` macro:\n\n```elixir\ndefmodule CallingAnother do\n  use Charms\n  alias Charms.Term\n\n  defm callee_function(arg :: Term.t()) :: Term.t() do\n    func.return(arg)\n  end\n\n  defm caller_function(env, arg :: Term.t()) :: Term.t() do\n    fun = call callee_function(arg) :: Term.t()\n    fun = callee_function(arg)\n    fun = call HaveFun.my_fun(env, fun, fun) :: Term.t()\n    HaveFun.my_fun(env, fun, fun)\n  end\nend\n\nCallingAnother.caller_function(\"hello\")\n```\n\nThe `call` macro is essential when you want to declare the return type of an expression.","ref":"programming-with-charms.html#calling-other-functions"},{"type":"extras","title":"Using Pointers and Memory Operations - Tutorial: Programming with Charms","doc":"Charms provides low-level memory operations through the `Pointer` module:\n\n```elixir\ndefmodule Dangerous do\n  use Charms\n  alias Charms.{Term, Pointer}\n\n  defm use_a_pointer(env, value :: Term.t()) :: Term.t() do\n    ptr = Pointer.allocate(Term.t())\n    Pointer.store(value, ptr)\n    Pointer.load(Term.t(), ptr)\n  end\nend\n\nDangerous.use_a_pointer(1)\n```\n\nYes, Charms **can** crash BEAM. We hope that as the project evolves, its safety features will improve.","ref":"programming-with-charms.html#using-pointers-and-memory-operations"},{"type":"extras","title":"Native Types and ENIF Functions - Tutorial: Programming with Charms","doc":"Charms provides several native types. Here are some of the key types:\n\n* `Term.t()`: Represents an Erlang term\n* `Pointer.t()`: Represents a pointer\n* `Env.t()`: Represents the BEAM environment\n* `i32()`, `i64()`: Represent integer types\n\nYou can make use of Erlang's ENIF functions to work with these types. Let's define a native function that specifies types for both arguments and the return value:\n\n```elixir\ndefmodule PlayWithErlangRuntime do\n  use Charms\n  alias Charms.{Env, Term, Pointer}\n\n  defm do_add(a :: i32(), b :: i32()) :: i32() do\n    a + b\n  end\n\n  defm add(env :: Env.t(), i) :: Term.t() do\n    i_ptr = Pointer.allocate(i32())\n    enif_get_int(env, i, i_ptr)\n    i = Pointer.load(i32(), i_ptr)\n    sum = do_add(i, i)\n    enif_make_int(env, sum)\n  end\nend\n\nPlayWithErlangRuntime.add(1)\n```\n\n`add/1` is the only function exported to Elixir in this module, which means you cannot invoke the `do_add/2` function from Elixir. Use `function_exported?/3` to verify this:\n\n```elixir\nfunction_exported?(MyModule1, :do_add, 2)\n```","ref":"programming-with-charms.html#native-types-and-enif-functions"},{"type":"extras","title":"Control Flow - Tutorial: Programming with Charms","doc":"Charms supports conditional statements and loops, which are vital for implementing logic in your native code.","ref":"programming-with-charms.html#control-flow"},{"type":"extras","title":"If-Else Control - Tutorial: Programming with Charms","doc":"For conditional logic, you can use `if` expressions. Below is an example that compares two values and returns the lesser one:\n\n```elixir\ndefmodule ControlFlowIf do\n  use Charms\n  alias Charms.Term\n\n  defm if_example(env, a, b) :: Term.t() do\n    if enif_compare(a, b) < 0 do\n      a\n    else\n      b\n    end\n  end\nend\n\nControlFlowIf.if_example(1, 2)\n```\n\nIn this function, we utilize `enif_compare/2` to determine which value is smaller, permitting a straightforward control flow. Notably, in Charms, both branches of the `if` must return values of the same type.","ref":"programming-with-charms.html#if-else-control"},{"type":"extras","title":"While Loops - Tutorial: Programming with Charms","doc":"You can implement loops using `while` constructs. Here's an example that iterates over a list to find the minimum value:\n\n```elixir\ndefmodule ControlFlowWhile do\n  use Charms\n  alias Charms.{Term, Pointer}\n\n  defm loop_example(env, list :: Term.t()) :: Term.t() do\n    movable_list_ptr = Pointer.allocate(Term.t())\n    Pointer.store(list, movable_list_ptr)\n    head_ptr = Pointer.allocate(Term.t())\n    min_ptr = Pointer.allocate(Term.t())\n\n    enif_get_list_cell(\n      env,\n      Pointer.load(Term.t(), movable_list_ptr),\n      min_ptr,\n      movable_list_ptr\n    )\n\n    while(\n      enif_get_list_cell(\n        env,\n        Pointer.load(Term.t(), movable_list_ptr),\n        head_ptr,\n        movable_list_ptr\n      ) > 0\n    ) do\n      min = Pointer.load(Term.t(), min_ptr)\n      head = Pointer.load(Term.t(), head_ptr)\n      min = ControlFlowIf.if_example(env, min, head)\n      Pointer.store(min, min_ptr)\n    end\n\n    Pointer.load(Term.t(), min_ptr)\n  end\nend\n\n[:a, :b, 1, 2, 3] |> Enum.shuffle() |> ControlFlowWhile.loop_example()\n```\n\nIn the `loop_example/2` function, we manage dynamic memory allocation for list traversal using pointers. The `while` checks if there are still elements to be processed in the list and updates the minimum value accordingly.\n\nEffectively employing control flow allows you to implement complex algorithms and logic directly in your Charms functions.","ref":"programming-with-charms.html#while-loops"},{"type":"extras","title":"Working with Constants - Tutorial: Programming with Charms","doc":"In addition to retrieving value types from Erlang terms, you can use the `const` macro to define constants with Elixir literals:\n\n```elixir\ndefmodule Constantly do\n  use Charms\n  alias Charms.Term\n\n  defm how_much(env) :: Term.t() do\n    zero = const 0 :: i32()\n    enif_make_int(env, zero + 100)\n  end\nend\n\nConstantly.how_much()\n```","ref":"programming-with-charms.html#working-with-constants"},{"type":"extras","title":"Error Handling - Tutorial: Programming with Charms","doc":"You can raise exceptions by creating an Erlang exception term using `enif_raise_exception`, as demonstrated below. Unlike Elixir's `raise`, which introduces special control flow, `enif_raise_exception` does not alter the control flow in the same way when raising exceptions.\n\n```elixir\ndefmodule Exceptional do\n  use Charms\n  alias Charms.{Term, Pointer}\n\n  @err_t %ArgumentError{message: \"not an int\"}\n  @err_v %ArgumentError{message: \"num too small\"}\n\n  defm raise_it(env, i :: Term.t()) :: Term.t() do\n    i_ptr = Pointer.allocate(i32())\n\n    if enif_get_int(env, i, i_ptr) != 0 do\n      if Pointer.load(i32(), i_ptr) > 0 do\n        i\n      else\n        enif_raise_exception(env, @err_v)\n      end\n    else\n      enif_raise_exception(env, @err_t)\n    end\n  end\nend\n\ntry do\n  Exceptional.raise_it(:a)\nrescue\n  e -> IO.inspect(e)\nend\n\ntry do\n  Exceptional.raise_it(0)\nrescue\n  e -> IO.inspect(e)\nend\n\nExceptional.raise_it(1)\n```\n\nIn this example, we employ a technique using `@err_t` and `@err_v`. In Charms, Elixir's module attributes are serialized and compiled as native values transparently.","ref":"programming-with-charms.html#error-handling"},{"type":"extras","title":"Conclusion - Tutorial: Programming with Charms","doc":"This tutorial provides an overview of the basics of programming with the Charms DSL, which lets you write native code that can be JIT-compiled while still utilizing a familiar Elixir-like syntax. For more advanced features and additional fun, be sure to consult the Charms documentation.","ref":"programming-with-charms.html#conclusion"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}